#!/usr/bin/env python

"""
rm.py - Rate Monotonic scheduler

RmPriorityQueue: priority queue that prioritizes by period
RmScheduler: scheduling algorithm that executes RM
"""

import json
import sys

from taskset import *
from scheduleralgorithm import *
from schedule import ScheduleInterval, Schedule
from display import SchedulingDisplay

class RmPriorityQueue(PriorityQueue):
    def __init__(self, jobReleaseDict):
        """
        Creates a priority queue of jobs ordered by period.
        """
        PriorityQueue.__init__(self, jobReleaseDict)

    def _sortQueue(self):
        # RM orders by period
        self.jobs.sort(key = lambda x: (x.task.period, x.task.id, x.id))

    def _findFirst(self, t):
        """
        Returns the index of the highest-priority job released at or before t,
        or -1 if the queue is empty or if all remaining jobs are released after t.
        """
        if self.isEmpty():
            return -1

        currentJobs = [(i, job) for (i,job) in enumerate(self.jobs) if job.releaseTime <= t]
        if len(currentJobs) == 0:
            return -1

        currentJobs.sort(key = lambda x: (x[1].task.period, x[1].task.id, x[1].id))
        return currentJobs[0][0] # get the index from the tuple in the 0th position

    def popNextJob(self, t):
        """
        Removes and returns the highest-priority job of those released at or after t,
        or None if no jobs are released at or after t.
        """
        laterJobs = [(i, job) for (i,job) in enumerate(self.jobs) if job.releaseTime >= t]

        if len(laterJobs) == 0:
            return None

        laterJobs.sort(key = lambda x: (x[1].releaseTime, x[1].task.period, x[1].task.id))
        return self.jobs.pop(laterJobs[0][0]) # get the index from the tuple in the 0th position

    def popPreemptingJob(self, t, job):
        """
        Removes and returns the job that will preempt job 'job' after time 't', or None
        if no such preemption will occur (i.e., if no higher-priority jobs
        are released before job 'job' will finish executing).

        t: the time after which a preemption may occur
        job: the job that is executing at time 't', and which may be preempted
        """
        hpJobs = [(i, j) for (i,j) in enumerate(self.jobs) if \
                  ((j.task.period < job.task.period or (j.task.period == job.task.period and j.task.id < job.task.id)) and \
                    j.releaseTime > t and j.releaseTime < t + job.remainingTime)]

        if len(hpJobs) == 0:
            return None

        hpJobs.sort(key = lambda x: (x[1].releaseTime, x[1].task.period, x[1].task.id))
        return self.jobs.pop(hpJobs[0][0]) # get the index from the tuple in the 0th position

class RmScheduler(SchedulerAlgorithm):
    def __init__(self, taskSet):
        SchedulerAlgorithm.__init__(self, taskSet)

    def buildSchedule(self, startTime, endTime):
        self._buildPriorityQueue(RmPriorityQueue)

        time = 0.0
        self.schedule.startTime = time

        previousJob = None
        didPreemptPrevious = False

        # Loop until the priority queue is empty, executing jobs preemptively in RM order
        while not self.priorityQueue.isEmpty():
            # Make a scheduling decision resulting in an interval
            interval, newJob = self._makeSchedulingDecision(time, previousJob)

            nextTime = interval.startTime
            didPreemptPrevious = interval.didPreemptPrevious

            # If the previous interval wasn't idle, execute the previous job
            # until the start of the new interval
            if previousJob is not None:
                previousJob.execute(interval.startTime - time)

                # If a preemption occurred, put the previous job back in the
                # priority queue
                if didPreemptPrevious and not previousJob.isCompleted():
                    self.priorityQueue.addJob(previousJob)

            # Add the interval to the schedule
            self.schedule.addInterval(interval)

            # Update the time and job
            time = nextTime
            previousJob = newJob

        # If there is still a previous job, complete it and update the time
        if previousJob is not None:
            time += previousJob.remainingTime
            previousJob.executeToCompletion()

        # Add the final idle interval
        finalInterval = ScheduleInterval()
        finalInterval.intialize(time, None, False)
        self.schedule.addInterval(finalInterval)

        # Post-process the intervals to set the end time and whether the job completed
        latestDeadline = max([job.deadline for job in self.taskSet.jobs])
        endTime = max(time + 1.0, latestDeadline, float(endTime))
        self.schedule.postProcessIntervals(endTime)

        return self.schedule

    def _makeSchedulingDecision(self, t, previousJob):
        """
        Makes a scheduling decision after time t.

        t: the beginning of the previous time interval, if one exists (or 0 otherwise)
        previousJob: the job that was previously executing, and will either complete or be preempted

        returns: (ScheduleInterval instance, Job instance of new job to execute)
        """
        if previousJob is None:
            # If there was no previous job, the last interval was an idle one
            newJob = self.priorityQueue.popNextJob(t)
            nextTime = newJob.releaseTime
            doesPreempt = False
        else:
            # If a job will preempt the previous job (due to the new job's
            # release), find that job
            preemptingJob = self.priorityQueue.popPreemptingJob(t, previousJob)

            if (preemptingJob is not None) and (previousJob.remainingTime == preemptingJob.releaseTime - t):
                # No preemption occurs
                intervalDuration = previousJob.remainingTime
                nextTime = t + intervalDuration
                doesPreempt = False

                newJob = preemptingJob
            elif (preemptingJob is None):
                # No preemption occurs
                intervalDuration = previousJob.remainingTime
                nextTime = t + intervalDuration
                doesPreempt = False

                # Choose a new job to execute (if None, this will be an idle interval)
                newJob = self.priorityQueue.popFirst(nextTime)
            else:
                # A preemption occurs
                nextTime = preemptingJob.releaseTime
                doesPreempt = True

                newJob = preemptingJob

        # Build the schedule interval
        interval = ScheduleInterval()
        interval.intialize(nextTime, newJob, doesPreempt)

        return interval, newJob

if __name__ == "__main__":
    if len(sys.argv) > 1:
        file_path = sys.argv[1]
    else:
        file_path = "tasksets/hwp1_test1.json"

    with open(file_path) as json_data:
        data = json.load(json_data)

    taskSet = TaskSet(data)

    taskSet.printTasks()
    taskSet.printJobs()

    rm = RmScheduler(taskSet)
    schedule = rm.buildSchedule(0, 6)

    schedule.printIntervals(displayIdle=True)

    print("\n// Validating the schedule:")
    schedule.checkWcets()
    schedule.checkFeasibility()

    display = SchedulingDisplay(width=800, height=480, fps=33, scheduleData=schedule)
    display.run()